#!/usr/bin/env python3

#
# Dun Dut Ut
#
# A Winter Chip IX Wild Chip entry
# by rainwarrior, 2/04/2014
#
# This "wild chip" entry implements an array
# of tuned ringing filters, stimulated by bursts
# of 1-bit noise. The final signal includes
# a simple echo as well.
#

import sys
assert sys.version_info[0] >= 3, "Python 3 required."

import array
import wave
import random
import math

# output configuration
# ====================

filename = "DUNDUTUT.WAV"
#sps = 44100
sps = 96000 # higher samplerate eliminates ringing at nyquist freq
fps = 60
spf = int(sps/fps)

# tuning setup and data enumerations
# ==================================

root = 146.832 # D in Hz
octaves = 3
decay = 0.05 * sps # time for instrument to decay to 37%
echo_time = 12 * spf # echo buffer length
echo_feedback = 0.4 # echo feedback strength
echo_level = 0.8 # total echo level
pan_width = 0.8 # stereo panning width
amp_scale = 3.0 # higher pitches are quieter, this compensates linearly
gain = 0.6 # overall gain

scale = [
    1.0  / 1.0, # Do  unison
    9.0  / 8.0, # Eo  major second
    7.0  / 6.0, # Fh  septimal minor third
    6.0  / 5.0, # Fo  minor third
    5.0  / 4.0, # Fs  major third
    4.0  / 3.0, # Go  fourth
    3.0  / 2.0, # Ao  fifth
    5.0  / 3.0, # Bo  major sixth
    7.0  / 4.0, # Ch  septimal minor seventh
    16.0 / 9.0, # Co  minor seventh
    15.0 / 8.0, # Cs  major seventh
    ]

COMMAND = 256 - ((len(scale) * octaves) + 4)
AMP   = COMMAND # next value changes amplitude (0 to 255)
BURST = AMP+1   # next value changes noise burst length (0 to 255)
END   = BURST+1 # end the tune
NOTE  = END+1
Do0   = NOTE
Eo0   = Do0+1
Fh0   = Eo0+1
Fo0   = Fh0+1
Fs0   = Fo0+1
Go0   = Fs0+1
Ao0   = Go0+1
Bo0   = Ao0+1
Ch0   = Bo0+1
Co0   = Ch0+1
Cs0   = Co0+1
Do1   = Cs0+1
Eo1   = Do1+1
Fh1   = Eo1+1
Fo1   = Fh1+1
Fs1   = Fo1+1
Go1   = Fs1+1
Ao1   = Go1+1
Bo1   = Ao1+1
Ch1   = Bo1+1
Co1   = Ch1+1
Cs1   = Co1+1
Do2   = Cs1+1
Eo2   = Do2+1
Fh2   = Eo2+1
Fo2   = Fh2+1
Fs2   = Fo2+1
Go2   = Fs2+1
Ao2   = Go2+1
Bo2   = Ao2+1
Ch2   = Bo2+1
Co2   = Ch2+1
Cs2   = Co2+1
Do3   = Cs2+1

unit_count = (octaves*len(scale)) + 1

# song data
# =========

songlen = 30.0 # seconds
song = [
    AMP, 128, BURST, 100,
    15,
    # test
    #Do0,6,Eo0,6,Fh0,6,Fo0,6,Fs0,6,Go0,6,Ao0,6,Bo0,6,Ch0,6,Co0,6,Cs0,6,
    #Do1,6,Eo1,6,Fh1,6,Fo1,6,Fs1,6,Go1,6,Ao1,6,Bo1,6,Ch1,6,Co1,6,Cs1,6,
    #Do2,6,Eo2,6,Fh2,6,Fo2,6,Fs2,6,Go2,6,Ao2,6,Bo2,6,Ch2,6,Co2,6,Cs2,6,
    #Do3,6,120,
    #AMP,64,Do1,60, AMP,255,Do1,60,
    #AMP,128,BURST,1,Ao0,60, BURST,32,Ao0,60, BURST,128,Ao0,60, BURST,255,Ao0,60,
    # song verse 1
    BURST,100,Ao0,8,Do1,8,Ch1,8,Go0,8,Eo0,8,8,Go0,8,BURST,255,Ao0,8,
    BURST,100,Ao0,8,Do1,8,Ch1,8,Go0,8,Eo0,8,8,Go0,8,BURST,255,Ao0,8,
    BURST,100,Ch1,8,Do1,8,Fo1,8,Do1,8,Ch0,8,Ao0,8,BURST,255,Ch0,8,Do1,8,
    BURST,100,Ch1,8,Do1,8,Fo1,8,Do1,8,Ch0,8,Ao0,8,BURST,255,Ch0,8,Do1,8,
    BURST,100,Ao0,Go1,8,Fo1,8,Do1,8,8,Do0,Fh1,8,Do1,8,Ch0,8,8,
    BURST,100,Bo0,Do2,8,Co1,8,Go1,8,Ao1,8,Go0,Co1,8,Go1,8,Fh1,8,8,
    BURST,100,Co0,Do1,8,Eo1,8,Fs1,8,Go1,8,Go0,Ao1,8,Bo1,8,Ch1,8,Cs1,8,
    BURST,100,Do0,Do2,8,Ao1,8,Ch1,8,Go1,8,Eo0,Eo1,8,8,Fh0,Go1,8,Ao1,8,
    BURST,100,Go0,Do2,8,Ao1,8,Ch1,8,Go1,8,Bo0,Eo1,8,8,Go1,8,Ao1,8,
    BURST,100,Co0,Ch1,8,Co1,8,Ao0,Cs1,8,Do2,8,Ao0,Eo2,8,Go2,8,Ao0,Fh2,8,Co1,8,
    BURST,100,Go0,Eo2,8,Eo2,8,Go0,Fh2,8,Eo2,8,Fo0,Do2,8,Co1,8,Eo0,Eo2,8,Go1,8,
    # song verse 2
    BURST,100,Ao0,8,Do1,8,Ch1,8,Go0,8,Eo0,8,8,Go0,8,BURST,255,Ao0,8,
    BURST,100,Ao0,8,Do1,8,Ch1,8,Go0,8,Eo0,8,8,Go0,8,BURST,255,Ao0,8,
    BURST,100,Ch1,8,Do1,8,Fo1,8,Do1,8,Ch0,8,Ao0,8,BURST,255,Ch0,8,Do1,8,
    BURST,100,Ch1,8,Do1,8,Fo1,8,Do1,8,Ch0,8,Ao0,8,BURST,255,Ch0,8,Do1,8,
    BURST,100,Ao0,Go1,8,Fo1,8,Do1,8,8,Do0,Fh1,8,Do1,8,Ch0,8,8,
    BURST,100,Bo0,Do2,8,Co1,8,Go1,8,Bo1,8,Go0,Co1,8,Eo1,8,Fh1,8,Fo1,8,
    BURST,100,Co0,Fs1,8,Go1,8,Do1,8,Ch0,8,Go0,Do1,8,Go1,8,Ao1,8,Co0,8,
    BURST,100,Do0,Ch1,8,Bo1,8,Ao1,8,Go1,8,Eo0,Ao1,8,Eo2,8,Fh0,Do2,8,Go1,8,
    BURST,100,Go0,Ao1,8,Ch1,8,Do2,8,Go2,8,Bo0,Fo2,8,Fh2,8,Eo2,8,Do2,8,
    BURST,100,Co0,Co1,8,8,Ch0,Do2,8,Co1,8,Bo0,Ao1,8,8,Ao0,Ch1,8,Ao1,8,
    BURST,100,Go0,Go1,8,8,Go0,Ao1,8,Go1,8,Fo0,Eo1,8,Do1,8,Eo0,Ch0,8,Go0,8,
    # gliss ending
    Do0,3,Eo0,3,Fh0,3,Fo0,3,Fs0,3,Go0,3,Ao0,3,Bo0,3,Ch0,3,Co0,3,Cs0,3,
    Do1,3,Eo1,3,Fh1,3,Fo1,3,Fs1,3,Go1,3,Ao1,3,Bo1,3,Ch1,3,Co1,3,Cs1,3,
    Do2,3,Eo2,3,Fh2,3,Fo2,3,Fs2,3,Go2,3,Ao2,3,Bo2,3,Ch2,3,Co2,3,Cs2,3,
    BURST,80,Do3,3,120,120,
    END
    ]

# implementation
# ==============

# echo

echo_buffer = array.array("f",[0.0]*echo_time)
echo_pos = 0

def tick_echo(x):
    global echo_buffer
    global echo_pos
    old = echo_buffer[echo_pos] * echo_feedback
    echo_buffer[echo_pos] = old + x
    echo_pos += 1
    if echo_pos >= echo_time:
        echo_pos = 0
    return old

# noise (1-bit LFSR based on NES)

noise = 1
noise_amp = 0.0

def tick_noise():
    global noise
    feedback = (noise & 1) ^ ((noise >> 1) & 1)
    noise = (noise >> 1) | (feedback << 14)
    if feedback != 0:
        return noise_amp
    else:
        return 0.0

# ringing filter units

class Unit:
    def __init__(self,hz,pan,amp):
        # sine law panning
        self.ampl = math.cos(pan*math.pi/2.0) * amp
        self.ampr = math.sin(pan*math.pi/2.0) * amp
        # filter constants
        self.fc = 2.0 * math.sin(math.pi * hz / sps)
        self.tc = math.exp(-1.0 / decay)
        # delay lines
        self.da = 0.0
        self.db = 0.0
    def tick(self,x):
        self.da += x
        output = self.da * self.fc + self.db
        self.db = output
        self.da = (self.da * self.tc) - (output * self.fc)
        return (output*self.ampl,output*self.ampr)

units = []

def setup_units():
    global units
    base_freq = root
    scale_pos = 0
    pan = 0.5 - (pan_width / 2)
    for i in range(0,unit_count):
        freq = base_freq * scale[scale_pos]
        amp = 1.0 + (((amp_scale-1.0)*i)/(unit_count-1))
        unit = Unit(freq,pan,amp)
        units.append(unit)
        scale_pos += 1
        if scale_pos >= len(scale):
            scale_pos = 0
            base_freq *= 2.0
        pan += (pan_width / (unit_count-1))

setup_units()

def tick_unit(unit,x):
    global units
    return units[unit].tick(x)

# mixer

# - reads and executes song data
# - responsible for ticking all other units

amp = 0.0 # current AMP setting
burst_len = 1 # samples per burst
note_on = [False]*(1+(len(scale)*octaves))

frame_time = 0 # time left in frame
burst_time = 0 # time left in burst
song_wait = 0 # frames to wait for next song read
song_pos = 0 # position in song data
ended = False # end of song reached

rendered_second = 0
samples_rendered = 0

def tick_mix():
    global noise_amp
    global amp
    global burst_len
    global note_on
    global frame_time
    global burst_time
    global song_wait
    global song_pos
    global ended
    global rendered_second
    global samples_rendered
    # read song
    if not ended:
        if (frame_time <= 0):
            # time indicator
            samples_rendered += spf
            if int(samples_rendered / sps) > rendered_second:
                rendered_second = int(samples_rendered / sps)
                print("Rendering %ds..."%rendered_second)
            # tick frame
            frame_time = spf
            note_on = [False]*(1+(len(scale)*octaves))
            while song_wait <= 0 and not ended:
                # read a byte from the song data
                if song_pos >= len(song):
                    ended = True
                    break
                c = song[song_pos]
                song_pos += 1
                if c < COMMAND:
                    song_wait = c
                else:
                    if c >= NOTE:
                        note_on[c-NOTE] = True
                        burst_time = burst_len
                    elif c == AMP:
                        v = song[song_pos]
                        song_pos += 1
                        amp = (2.0 * v) / 255.0
                        noise_amp = amp / burst_len
                    elif c == BURST:
                        v = song[song_pos]
                        song_pos += 1
                        burst_len = v
                        noise_amp = amp / burst_len
                    elif c == END:
                        ended = True
            song_wait -= 1
        frame_time -= 1
    # tick noise
    n = 0.0
    if burst_time > 0:
        n = tick_noise()
        burst_time -= 1
    # tick and mix units
    snd_l = 0.0
    snd_r = 0.0
    for i in range(0,unit_count):
        x = 0.0
        if note_on[i]:
            x = n
        (ul,ur) = tick_unit(i,x)
        snd_l += ul
        snd_r += ur
    # echo
    echo = tick_echo((snd_l+snd_r)*0.5)
    snd_l += echo * echo_level
    snd_r += echo * echo_level
    # output
    return (snd_l*gain,snd_r*gain)

# main
# ====

def render():
    d = bytearray()
    wavelen = int(sps * songlen)
    for i in range(0,wavelen):
        (a,b) = tick_mix()
        # pack as 16 bit signed samples
        a = max(-32767,min(32767,int(a*32767.0)))
        b = max(-32767,min(32767,int(b*32767.0)))
        d.append(a & 255)
        d.append((a >> 8) & 255)
        d.append(b & 255)
        d.append((b >> 8) & 255)
    # save wave
    w = wave.open(filename,"wb")
    w.setnchannels(2)
    w.setsampwidth(2)
    w.setframerate(sps)
    w.writeframes(d)
    w.close()
    print("generated: %s" % filename)

# execute

if __name__ == "__main__":
    render()

# end of file
